/* 
 * Base64 decode
 *
 * Copyright 1996-2002 netFluid Technology Ltd
 * http://www.nfluid.co.uk
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 * $Revision: 1.1.1.1 $  $Date: 2003/01/10 23:15:54 $
 * ---------------------------------------------------------------------------
 */
#include <phlib/xl_config.h>
#include <phlib/xl_types.h>
#include <phlib/xl_base64_decode.h>

/* Need to define machine word byte ordering */
#if defined(AC_C_BIGENDIAN)
#  define S1 1
#  define S2 2
#  define S3 3
#else
#  define S1 2
#  define S2 1
#  define S3 0
#endif

/* This is the decode table. It's mostly redundant, 
 * but decoding is faster this way
 */

static const unsigned char base64_decode_table[0x100] = 
{
  0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, /* 0x00 - 0x3F */
  0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
  0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
  0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
  0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
  0xff, 0xff, 0xff, 0x3E, 0xff, 0xff, 0xff, 0x3F,
  0x34, 0x35, 0x36, 0x37, 0x38, 0x39, 0x3A, 0x3B,
  0x3C, 0x3D, 0xff, 0xff, 0xff, 0x00, 0xff, 0xff,
  0xff, 0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, /* 0x40 - 0x7F */
  0x07, 0x08, 0x09, 0x0A, 0x0B, 0x0C, 0x0D, 0x0E,
  0x0F, 0x10, 0x11, 0x12, 0x13, 0x14, 0x15, 0x16,
  0x17, 0x18, 0x19, 0xff, 0xff, 0xff, 0xff, 0xff, 
  0xff, 0x1A, 0x1B, 0x1C, 0x1D, 0x1E, 0x1F, 0x20,
  0x21, 0x22, 0x23, 0x24, 0x25, 0x26, 0x27, 0x28,
  0x29, 0x2A, 0x2B, 0x2C, 0x2D, 0x2E, 0x2F, 0x30,
  0x31, 0x32, 0x33, 0xff, 0xff, 0xff, 0xff, 0xff, 
  0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, /* 0x80 - 0xBF */
  0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
  0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
  0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
  0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
  0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
  0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
  0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
  0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, /* 0xC0 - 0xFF */
  0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
  0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
  0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
  0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
  0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
  0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
  0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff
};

/* ------------------------------------------------------------------------- */
/* Decode base 64 encoded stuff. Messy, but covers RFC1521 fairly well 
 * compared to simpler versions 
 */

size_t 
xl_Base64_decode( void *dst,const void *src,size_t size )
{
  size_t decoded=0;
  register const unsigned char *in=(const unsigned char*)src;
  register unsigned char *out=(unsigned char*)dst;
   
  /* Repeat while we still have enough input data 
     for another sequence */
  while( size > 3 ) {
    unsigned char store[4];
	
    /* Skip whitespace */
    while( base64_decode_table[ *in ] & 0xC0 ) {
      in++;
      if( (--size) < 4 ) return decoded;
    }

    /* Grab next 4 coded bytes */
    store[0] = *(in++);
    store[1] = *(in++);
    store[2] = *(in++);
    store[3] = *(in++);
    size -= 4;

    /* Begin decode */
    if( (store[0] == '=' ) || (store[1] == '=' ) ) {
      /* Error - illegal sequence - skip it */
    }
    else if (store[2] == '=') {
      /* Possible end sequence */
      if (store[3] == '=') {
	/* End sequence - 8 bits */
	unsigned long bucket=0;
	bucket  |= base64_decode_table[store[0]];
	bucket <<= 6;
	bucket  |= base64_decode_table[store[1]];
	bucket <<= 12;

	/* Append to output */
	*(out++) = ((unsigned char*)&bucket)[S1];
	decoded += 1;
	break;
      }
      /* Error - illegal sequence - skip it */
    }
    else if (store[3] == '=') {
      /* End sequence - 16 bits */
      unsigned long bucket = 0;
      bucket  |= base64_decode_table[store[0]];
      bucket <<= 6;
      bucket  |= base64_decode_table[store[1]];
      bucket <<= 6;
      bucket  |= base64_decode_table[store[2]];
      bucket <<= 6;

      /* Append to output */
      *(out++) = ((unsigned char*)&bucket)[S1];
      *(out++) = ((unsigned char*)&bucket)[S2];
      decoded += 2;
      break;
    }
    else {
      /* Normal sequence */
      unsigned long bucket=0;
      bucket  |= base64_decode_table[store[0]];
      bucket <<= 6;
      bucket  |= base64_decode_table[store[1]];
      bucket <<= 6;
      bucket  |= base64_decode_table[store[2]];
      bucket <<= 6;
      bucket  |= base64_decode_table[store[3]];

      /* Append to output */
      *(out++) = ((unsigned char*)&bucket)[S1];
      *(out++) = ((unsigned char*)&bucket)[S2];
      *(out++) = ((unsigned char*)&bucket)[S3];
      decoded += 3;
    }
  } 

  /* Done */
  return decoded;
}

/* ------------------------------------------------------------------------- */
/* end */
