/* 
 * This is a lazy date/time creation module.  It calculates the date
 * using system call 'localtime' and from that moment on compares the date
 * in seconds since 1970 and decides what to change...
 * ------------------------------------------------------------------------
 *
 * Copyright 1996-2002 netFluid Technology Ltd
 * http://www.nfluid.co.uk
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 * $Revision: 1.1.1.1 $  $Date: 2003/01/10 23:16:00 $
 * ---------------------------------------------------------------------- 
 */

#include <phlib/xl_config.h>
#include <unistd.h>
#include <sys/time.h>

#if TIME_WITH_SYS_TIME
# include <sys/time.h>
# include <time.h>
#else
# if HAVE_SYS_TIME_H
#  include <sys/time.h>
# else
#  include <time.h>
# endif
#endif

#if STDC_HEADERS
# include <string.h>
#else
# ifndef HAVE_MEMCPY
#  define memcpy(d, s, n) bcopy ((s), (d), (n))
#  define memmove(d, s, n) bcopy ((s), (d), (n))
# endif
#endif

#include <phlib/xl_types.h>
#include <phlib/xl_date.h>

static struct tm curr_time;
static time_t    last_sec;

static uchar_t timestr[] = "00:00:00";
static uchar_t datestr[] = "0000/00/00";

#define H0_IX 0
#define H1_IX 1
#define N0_IX 3
#define N1_IX 4
#define S0_IX 6
#define S1_IX 7

#define Y0_IX 0
#define Y1_IX 1
#define Y2_IX 2
#define Y3_IX 3
#define M0_IX 5
#define M1_IX 6
#define D0_IX 8
#define D1_IX 9

#if 0  /* Not required */
static time_t tzone; /* Local Timezone adjustment */
#endif

/* Prototypes */
static void get_date();

/* ------------------------------------------------------------------------ */
/*
 * This function MUST be called to initialise the date/time system BEFORE
 * any calls to xl_get_date or xl_get_time.
 */
void
xl_init_date()
{
  last_sec = time( 0 ) - 1; /* -1 Hotwires the time so that diff > 0 */
  xl_date_resync( &last_sec );
}


/* ------------------------------------------------------------------------ */

void
xl_date_resync( time_t *last_sec )
{
  struct tm* ts;
  int    year;
  /* extern time_t timezone; */

  ts = localtime( last_sec ); 

  memcpy( &curr_time, ts, sizeof( struct tm ) );

  datestr[D0_IX] = (curr_time.tm_mday / 10) + '0';
  datestr[D1_IX] = (curr_time.tm_mday % 10) + '0';
      
  datestr[M0_IX] = ( (1 + curr_time.tm_mon) / 10 ) + '0';
  datestr[M1_IX] = ( (1 + curr_time.tm_mon) % 10 ) + '0';

  year = curr_time.tm_year + 1900;
  datestr[Y3_IX] = (year % 10) + '0'; year /= 10;
  datestr[Y2_IX] = (year % 10) + '0'; year /= 10;
  datestr[Y1_IX] = (year % 10) + '0'; year /= 10;
  datestr[Y0_IX] = (year % 10) + '0';

  timestr[S0_IX] = ( curr_time.tm_sec / 10 ) + '0';
  timestr[S1_IX] = ( curr_time.tm_sec % 10 ) + '0';
    
  timestr[N0_IX] = ( curr_time.tm_min / 10 ) + '0';
  timestr[N1_IX] = ( curr_time.tm_min % 10 ) + '0';

  timestr[H0_IX] = ( curr_time.tm_hour / 10 ) + '0';
  timestr[H1_IX] = ( curr_time.tm_hour % 10 ) + '0';
}

/* ------------------------------------------------------------------------ */
/*
 * Return the current date in yyyy/mm/dd format.
 *
 * If Entry flag is 1 then the date/time is re-calculated. Otherwise
 *  the last known date/time is used.
 *
 * Function returns a pointer to the date string.
 */
uchar_t *
xl_get_date( int flag )
{
  if( flag == 1 ) get_date();
  return datestr;
}

/* ------------------------------------------------------------------------ */
/*
 * Return the current time in hh:mm:ss format.
 *
 * If Entry flag is 1 then the date/time is re-calculated. Otherwise
 *  the last known date/time is used.
 *
 * Function returns a pointer to the time string
 */
uchar_t *
xl_get_time( int flag )
{
  if( flag == 1 ) get_date();
  return timestr;
}


/* ------------------------------------------------------------------------ */
/* 
 * This function will fail if it is called less than once a year!  This is
 * because it only considers the year in which it was last called for leap
 * year checking.
 *
 * This is a big function, :-(   But this way it should be fast. ;-)
 */

static void
get_date()
{
  time_t now;
  int    diff;
  int    year;
  int    days = 365;
  int    dd,hh,mm,ss,MM;
  struct timeval tv;

  MM = 0;
  gettimeofday( &tv, NULL );
  now = tv.tv_sec;

#if 0 /* Localtime dones this anyway */
  now += tzone; /* Adjust for timezone and BST if applicable */
#endif

  diff = now - last_sec;
  last_sec = now;
  
  if( diff > 0 ) {
    
    year = curr_time.tm_year + 1900;
    
    if((year % 4) == 0 && ((year % 100) != 0 || (year % 400) == 0)) days++;
    
    dd = diff / 86400; diff -= dd * 86400;
    hh = diff / 3600;  diff -= hh * 3600;
    mm = diff / 60;    diff -= mm * 60;
    ss = diff;
    
    ss += curr_time.tm_sec;                      /* Inc seconds          */
    mm += curr_time.tm_min  + ss / 60; ss %= 60; /* Inc Mins and rnd sec */
    hh += curr_time.tm_hour + mm / 60; mm %= 60; /* Inc hrs and rnd mins */
    
    dd += hh / 24;  hh %= 24;                    /* True day adjustment  */
    
    if( dd > 0 ) {  /* We've got to update the day... */
      
      curr_time.tm_wday = (curr_time.tm_wday + dd) % 7; /* Day of week   */
      curr_time.tm_yday += dd;                          /* Year day      */
      
      /* Calculate the month we are in from the day in the year          */
      dd = (curr_time.tm_yday % days ) + 1;
      
      switch( days ) {
      case 365: /* Non leap year */
	if( dd <= 31  ) { MM = 0;  curr_time.tm_mday = dd;       break; }
	if( dd <= 59  ) { MM = 1;  curr_time.tm_mday = dd - 31;  break; }
	if( dd <= 90  ) { MM = 2;  curr_time.tm_mday = dd - 59;  break; }
	if( dd <= 120 ) { MM = 3;  curr_time.tm_mday = dd - 90;  break; }
	if( dd <= 151 ) { MM = 4;  curr_time.tm_mday = dd - 120; break; }
	if( dd <= 181 ) { MM = 5;  curr_time.tm_mday = dd - 151; break; }
	if( dd <= 212 ) { MM = 6;  curr_time.tm_mday = dd - 181; break; }
	if( dd <= 243 ) { MM = 7;  curr_time.tm_mday = dd - 212; break; }
	if( dd <= 273 ) { MM = 8;  curr_time.tm_mday = dd - 243; break; }
	if( dd <= 304 ) { MM = 9;  curr_time.tm_mday = dd - 273; break; }
	if( dd <= 334 ) { MM = 10; curr_time.tm_mday = dd - 304; break; }
	if( dd <= 365 ) { MM = 11; curr_time.tm_mday = dd - 334; break; }
	break;
	
      case 366: /* Leap year     */
	if( dd <= 31  ) { MM = 0;  curr_time.tm_mday = dd;       break; }
	if( dd <= 60  ) { MM = 1;  curr_time.tm_mday = dd - 31;  break; }
	if( dd <= 91  ) { MM = 2;  curr_time.tm_mday = dd - 60;  break; }
	if( dd <= 121 ) { MM = 3;  curr_time.tm_mday = dd - 91;  break; }
	if( dd <= 152 ) { MM = 4;  curr_time.tm_mday = dd - 121; break; }
	if( dd <= 182 ) { MM = 5;  curr_time.tm_mday = dd - 152; break; }
	if( dd <= 213 ) { MM = 6;  curr_time.tm_mday = dd - 182; break; }
	if( dd <= 244 ) { MM = 7;  curr_time.tm_mday = dd - 213; break; }
	if( dd <= 274 ) { MM = 8;  curr_time.tm_mday = dd - 244; break; }
	if( dd <= 305 ) { MM = 9;  curr_time.tm_mday = dd - 274; break; }
	if( dd <= 335 ) { MM = 10; curr_time.tm_mday = dd - 305; break; }
	if( dd <= 366 ) { MM = 11; curr_time.tm_mday = dd - 335; break; }
	break;
      }
      
      /* The date has changed in some way, so adjust date string... */
      datestr[D0_IX] = (curr_time.tm_mday / 10) + '0';
      datestr[D1_IX] = (curr_time.tm_mday % 10) + '0';
      
      if( MM != curr_time.tm_mon ) {
	datestr[M0_IX] = ( MM / 10 ) + '0';
	datestr[M1_IX] = ( MM % 10 ) + '0';
	curr_time.tm_mon = MM;
      }
      
      if( (curr_time.tm_yday / days) != 0 ) {
	curr_time.tm_year += (curr_time.tm_yday / days);
	year = curr_time.tm_year + 1900;
	datestr[Y3_IX] = (year % 10) + '0'; year /= 10;
	datestr[Y2_IX] = (year % 10) + '0'; year /= 10;
	datestr[Y1_IX] = (year % 10) + '0'; year /= 10;
	datestr[Y0_IX] = (year % 10) + '0';
      }
    }     
    /* Now update the time strings as appropriate.... */
    curr_time.tm_sec = ss;
    timestr[S0_IX] = ( ss / 10 ) + '0';
    timestr[S1_IX] = ( ss % 10 ) + '0';
    
    if( mm != curr_time.tm_min ) {
      curr_time.tm_min = mm;
      timestr[N0_IX] = ( mm / 10 ) + '0';
      timestr[N1_IX] = ( mm % 10 ) + '0';
    }
    
    if( hh != curr_time.tm_hour ) {
      curr_time.tm_hour = hh;
      timestr[H0_IX] = ( hh / 10 ) + '0';
      timestr[H1_IX] = ( hh % 10 ) + '0';
    }
  }
}

/* ------------------------------------------------------------------------ */
/* end */
